/*                                                                        *
*    File is Copyright  2013 CSR Technology Inc.  All rights reserved.   *
*    No implied or expressed warranties are granted.                      *
*                                                                         *
*    This Software is protected by United States copyright laws and       *
*    international treaties.  You may not reverse engineer, decompile     *
*    or disassemble this Software.                                        *
*                                                                         *
*    WARNING:                                                             *
*    This Software contains CSR Technology Inc.s confidential and        *
*    proprietary information. UNAUTHORIZED COPYING, USE, DISTRIBUTION,    *
*    PUBLICATION, TRANSFER, SALE, RENTAL OR DISCLOSURE IS PROHIBITED      *
*    AND MAY RESULT IN SERIOUS LEGAL CONSEQUENCES.  Do not copy this      *
*    Software without CSR Technology, Inc.s  express written             *
*    permission. Use of any portion of the contents of this Software      *
*    is subject to and restricted by your signed written agreement with   *
*    CSR Technology, Inc.                                                 *
*                                                                         *
***************************************************************************
*
* MODULE:  ROM Patch Application
*
* FILENAME:  ROMPatchSettings.c
*
* DESCRIPTION:This file contains the source code for transfer and verification of the patch
*                         data.
*
***************************************************************************
*
*  Keywords for Perforce.  Do not modify.
*
*  $File: //core/tools/ROMPatchApplication/src/ROMPatchSettings.c $
*
*  $DateTime: 2015/09/28 18:35:52 $
*
*  $Revision: #13 $
*
***************************************************************************/

/**************************************************************************/
/*                              INCLUDES                                  */
/**************************************************************************/
#include "Abstraction/basicTypes.h"
#include "Abstraction/osAbstraction.h"
#include "Abstraction/fileOps.h"
#include "protocol.h"
#include "Abstraction/logs.h"
#include "ssbInterface.h"
#include "ROMPatchSettings.h"
#include "errorCodes.h"
#include "Abstraction/CommInterface.h"
#include "crc.h"
#include "Abstraction/string_abstraction.h"

/**************************************************************************/
/*                          GLOBALS                                       */
/**************************************************************************/
#define MAX_UINT16                  65535

/* Macro to read patch payload */
#define PATCH_READRAW(pRaw, type) \
      (type)(Patch_readRaw (pRaw, sizeof (type)));  pRaw += sizeof (type);

extern tInputConfig inputConfiguration;

/** Keeps track of the patch manager load state. */
tPMPatchState patchState;

/* This flag indicates if K32 patch is present in the input patch file */
 BOOL gK32PatchFlag; /* TRUE: Both ARM and K32 patches present */
                     /* FLASE: Only ARM patch present */

/**************************************************************************/
/*              STATIC FUNCTION PROTOTYPES                                */
/**************************************************************************/
static UNINT Read_PD2(FILE *fptr, UINT32 offset, UINT8 *data, UINT32 length);
static SINT32 ProcessPatchHeader(FILE *fptr, SINT32 offset, tPMPatchFileInfo *pPFInfo);
static SINT32 PM_ProcessFileInfo(FILE *fptr, SINT32 filePos, tPMPatchFileInfo *pPFInfo);
static SINT32 IsK32PatchAvailable(FILE *fptr);
static SINT32 LoadAndApplyPatch(FILE *fptr, SINT32 filePos, BOOL sendStartReq, BOOL sendEndReq);
static void InitializePatchState(BOOL validLoadInProgress);
static BOOL PM_FirstSegment(tPMPatchFileInfo *pi, tPMPatchFileState *pf);
static BOOL PM_NextSegment(tPMPatchFileState *pf);
static SINT32 GeneratePatchManagerStartRequest(UINT8 **ppData, LONG *plSize, LONG lNumBytesToWrite);
static SINT32 GeneratePatchManagerLoadRequest(UINT8 **ppData, LONG *plSize, UINT16 seqNum, LONG lNumBytesToWrite, UINT8 *pDataToWrite);
static SINT32 GeneratePatchManagerExitRequest(UINT8 **ppData, LONG *plSize, LONG lNumBytesToWrite);
static SINT32 writeSeqNumToStream(UINT8 *pData, SINT32 count, UINT16 seqNum);
static BOOL FindStartOfPatchRAMCode(tPMPatchFileInfo *pPFInfo, UINT8 *patchdata);
static UINT32 Patch_readRaw (UINT8 *pRaw, UINT32 size);
static UINT8 GetPowerMode(RcvrParamSS_2 *localParam);

/**************************************************************************/
/*                     GLOBAL FUNCTIONS                                   */
/**************************************************************************/
SINT32 ReadBinaryFile(SINT8* strPath, UINT8 **ppData, LONG *pSize)
{
    fileOpenOpsStruct fopenReq;
    fileRWOpsStruct readReq;
    SINT32 status = ERROR_NONE;

    PM_MemSet(&fopenReq, 0, sizeof(fileOpenOpsStruct));
    PM_MemSet(&readReq, 0, sizeof(fileRWOpsStruct));  

    if(strPath == "")
    {
        DebugPrint(("Error %d: Input File Name is Empty\n",ERROR_FILE_OPEN));
        SetErrorCode(ERROR_FILE_OPEN);
        return ERROR_FILE_OPEN;
    }

    /* Open the File */

    fopenReq.filePtr = NULL;

    PM_strcpy(fopenReq.fileName,strPath);
    PM_strcpy(fopenReq.mode,"rb");

    if(PM_OpenFile((void*)&fopenReq) == FAILURE)
    {
        DebugPrint(("Error %d: Unable To Open File %s\n",ERROR_FILE_OPEN, strPath));
        SetErrorCode(ERROR_FILE_OPEN);
        return ERROR_FILE_OPEN;
    }

    if(fopenReq.filePtr == NULL)
    {
        *pSize = 0;
        DebugPrint(("Error %d: Unable To Open File %s\n",ERROR_FILE_OPEN, strPath));
        SetErrorCode(ERROR_FILE_OPEN);
        return ERROR_FILE_OPEN;
    }

    /* Get the file size and request to read the complete file */

    readReq.filePtr = fopenReq.filePtr;
    readReq.numBytes = PM_GetFileSize(fopenReq.filePtr);
    readReq.offset1 = 0;
    readReq.ptr = PM_MemAlloc(sizeof(UINT8) * (readReq.numBytes));
    if(NULL == readReq.ptr )
    {
        DebugPrint(("Error %d: Error in memory allocation\n",ERROR_MEMALLOC));
        SetErrorCode(ERROR_MEMALLOC);
        return ERROR_MEMALLOC;
    }
    PM_ReadFile((void*)&readReq);

    /* Close the file */
    PM_CloseFile(fopenReq.filePtr);
    *pSize = readReq.numBytes;
    *ppData = readReq.ptr;

    return ERROR_NONE;
}

SINT32 ProcessPatchFile(SINT8 *patchFileName)
{
   SINT32 filePos;
   FILE *pd2 = NULL;
   gK32PatchFlag = FALSE;

   /* Open patch file for processing */
   pd2 = fopen(patchFileName,"rb");
   /* Check if patch file ptr is invalid */
   if (pd2 == NULL)
   {
       DebugPrint(("Error %d: Unable to open patch file!\n",ERROR_FILE_OPEN));
       SetErrorCode(ERROR_FILE_OPEN);
       return -1;
   }

   /* Determine if K32 patch exists in the input patch file */
   filePos = IsK32PatchAvailable(pd2);

   if(filePos == -1)
   {
      DebugPrint(("------Load ARM patch file-----\n"));
      LoadAndApplyPatch(pd2, 0, TRUE, TRUE);
   }
   else
   {
      DebugPrint(("------Load ARM section of patch file-----\n"));
      LoadAndApplyPatch(pd2, 0, TRUE, FALSE);
      DebugPrint(("------Load K32 section of patch file-----\n"));
      gK32PatchFlag = TRUE;
      LoadAndApplyPatch(pd2, filePos, FALSE, TRUE);
   }

   fclose(pd2);

   return filePos;
}

SINT32 RestoreRxSettings()
{
   SINT32 returnStatus = ERROR_NONE;

   if(NMEA_PROTOCOL == inputConfiguration.initialRxProtocol)
   {
      /* Based on following condition 115200 can not be supported. */
      if (inputConfiguration.initialRxBaudRate > MAX_UINT16)
      {
         DebugPrint(("FAIL: Cannot switch to NMEA baud greater than 57600\n"));
         returnStatus = FAILURE;
      }
      else
      {
         DebugPrint(("Switch to Initial Baud Rate %d\n",inputConfiguration.initialRxBaudRate));
         returnStatus = SwitchSsbToNmea((UINT16)inputConfiguration.initialRxBaudRate);
      }
   }
   else
   {
      /* Just need to change the baud rate */
      if(inputConfiguration.initialRxBaudRate != inputConfiguration.patchBaudRate)
      {
        /* Switch to Initial Baud Rate */
         UINT8 count= 0;
         DebugPrint(("Switch to Initial Baud Rate %d\n",inputConfiguration.initialRxBaudRate));
         while((count++ < MAX_NUM_TRIALS) && (returnStatus != ERROR_NONE_SSB_MSG_RCVD))
         {
            DebugPrint(("Trial %d\n",count));
            Comm_DevIOCtrl(UARTSetBaudrate,(void*)&inputConfiguration.patchBaudRate);
            DelayMs(DELAY_AFTER_BAUDRATECHANGE);
            SetSerialBitRate(inputConfiguration.initialRxBaudRate);
            returnStatus = ListenRx(response_recvSSB);
         }
         if(returnStatus == ERROR_NONE_SSB_MSG_RCVD)
            returnStatus = ERROR_NONE;
      }
   }
   return returnStatus;
}

/* The HandleLPModeScenario function sends the OSP command to get Nav Parameters. */
/* After going into full power mode, one may want to disable some OSP msgs by     */
/* sending MID166 commands to disable specific OSP msgs. Disabling msgs that are  */
/* not interrogated by this program will help reduce the msg processing while     */
/* waiting for a specific OSP response msg. MID255, MID64, and NavLib msgs are    */
/* good candidates to be disabled.                                                */
SINT32 HandleLPModeScenario()
{
    SINT32 returnStatus = ERROR_NONE;
    UINT8 count1,count=0;
    DebugPrint(("Trying To Get Receiver Power Mode Parameters\n"));

    /* Poll NAV Parameters to get the low power mode settings */
    DebugPrint(("Polling NAV \n"));
    PollNAVParameters();
    returnStatus = ListenRx(response_PollNavParams);
    /* DelayMs(DELAY_AFTER_RESET); */

    while( 
        ((ERROR_NACK_RCVD == returnStatus ) || (ERROR_MAX_NUMMSG_WAIT_EXPIRED == returnStatus ))
        && (count++ < 200)
        )
    {
        DebugPrint(("Retry Polling NAV \n"));
        PollNAVParameters();
        returnStatus = ListenRx(response_PollNavParams);
    }

    if(returnStatus == ERROR_NONE_SSB_MSG_RCVD)
        returnStatus = ERROR_NONE;

    if(returnStatus != ERROR_NONE) 
    {
        DebugPrint(("Failed to Get Power Mode Settings\n"));
        return returnStatus;
    }

    /* Take the backup of the received params */
    SavedRcvrParamSS2 = RcvrParamSS2;

    if(GetPowerMode(&SavedRcvrParamSS2) != 0)
    {
        DebugPrint(("Receiver in Low POWR \n"));

        for(count1=0;count1 <10;count1++)
        {
            DebugPrint(("Requesting Rx to move in FULL POWR\n"));

            EnterFullPower();

            PollNAVParameters();
            returnStatus = ListenRx(response_PollNavParams);
            count = 0;
            while( 
                ((ERROR_NACK_RCVD == returnStatus ) || (ERROR_MAX_NUMMSG_WAIT_EXPIRED == returnStatus ))
                && (count++ < 20)
                )
            {
                DebugPrint(("Polling NAV \n"));
                PollNAVParameters();
                returnStatus = ListenRx(response_PollNavParams);
            }

            if(returnStatus != ERROR_NONE)
                return returnStatus;

            if(GetPowerMode(&RcvrParamSS2) == 0)
            {
                DebugPrint(("Receiver moved to FULL POWR (SUCCESS)\n"));
                break;
            }
        }

        if(count == 2)
        {
            DebugPrint(("\nFAILURE! To Move in FULL POWR\n"));
            return ERROR_FAILED_TOCHANGE_PWRMODE;
        }
    }
    else
    {
        DebugPrint(("Receiver Started in FULL POWR MODE\n"));
    }
    return ERROR_NONE;
}

SINT32 RestoreLPModeSettings()
{
    SINT32 returnStatus = ERROR_NONE;
    UINT8 flag = GetPowerMode(&SavedRcvrParamSS2); 
    UINT8 count =0;
    /* 0 indicates full power , 1 or 2  indicates low power */

    if( flag != 0)
    {
        /* Receiver started in low power mode . Restore its settings */

        EnterLowPower(flag);
        PollNAVParameters();
        returnStatus = ListenRx(response_PollNavParams);
        if(GetPowerMode(&RcvrParamSS2) != flag)
        {
            /* if power mode has not changed , treat this as NACK Received */
            returnStatus = ERROR_NACK_RCVD;
        }

        while( 
            ((ERROR_NACK_RCVD == returnStatus ) || (ERROR_MAX_NUMMSG_WAIT_EXPIRED == returnStatus ))
            && (count++ < 20)
            )
        {
            /* Receiver started in low power mode . Restore its settings */
            EnterLowPower(flag);
            /* Poll again to verify if power mode is restored or not */
            PollNAVParameters();
            returnStatus = ListenRx(response_PollNavParams);
            if(GetPowerMode(&RcvrParamSS2) != flag)
            {
                /* if power mode has not changed , treat this as NACK Received */
                returnStatus = ERROR_NACK_RCVD;
            }
        }

        if(GetPowerMode(&RcvrParamSS2) != flag)
        {
            DebugPrint(("FAILURE! POWR MODE not restored \n"));
            return ERROR_FAILED_TOCHANGE_PWRMODE;
        }
        else
        {
            DebugPrint(("POWR MODE Restored (SUCCESS)\n"));
        }
    }
    else
    {
        DebugPrint(("Rx Started in FULL POWR, No POWR Mode Changes Required \n"));
    }

    return returnStatus;
}

/**************************************************************************/
/*                     STATIC FUNCTIONS                                   */
/**************************************************************************/
static UNINT Read_PD2(FILE *fptr, UINT32 offset, UINT8 *data, UINT32 length)
{
   UNINT result = 2; /* File not available */

   /* check if the file exists and the data pointer is not NULL */
   if((fptr != NULL) && (data != NULL))
   {
      if (0 == fseek(fptr, offset, SEEK_SET)) /* check if we can go to the place we need get to */
      {
         if (1 == fread(data,length,1,fptr))  /* check if the read succeeded */
         {
            result = 0; /* 0 - File successfully read */
         }
         else
         {
            result = 1;  /* 1 - File empty */
         }
      }
   }
   return result;
}

static SINT32 ProcessPatchHeader(FILE *fptr, SINT32 offset, tPMPatchFileInfo *pPFInfo)
{
   UINT8  tempbytes[PATCH_FILE_HEADERINFO_SIZE], *p = NULL;

   memset(tempbytes,0,sizeof(tempbytes));
   /* Read patch header information */
   if (0 == Read_PD2(fptr, offset, tempbytes, PATCH_FILE_HEADERINFO_SIZE))
   {
      p = tempbytes;
      pPFInfo->Head1 = *p++;
      pPFInfo->Head2 = *p++;
      /* Check to make sure patch file has proper start header */
      if ( (pPFInfo->Head1 == 'P') &&
           (pPFInfo->Head2 == 'D') )
      {
         pPFInfo->ChipId = p[0]<<8 | p[1];
         p += 2;
         pPFInfo->SiliconVersion = p[0]<<8 | p[1];
         p += 2;
         pPFInfo->ROMVersionCode = p[0]<<8 | p[1];
         p += 2;
         pPFInfo->PatchRevisionCode = p[0]<<8 | p[1];
         p += 2;
         pPFInfo->PatchDataBaseAddress = p[0]<<24 | p[1]<<16 | p[2]<<8 | p[3];
         p += 4;
         pPFInfo->PatchDataLength = p[0]<<8 | p[1];
         p += 2;

         if (pPFInfo->ROMVersionCode >= ROMVERSIONCODE_USE_CRC32)
         {
            gCRCType = CRC_32;
            gPatchFileCRCSize = 4;
         }
         else
         {
            gCRCType = CRC_16;
            gPatchFileCRCSize = 2;
         }
      }
      else
      {
         DebugPrint(("FAILURE! Improper start of Patch Header!\n"));
         return -1;
      }
   }
   else
   {
      DebugPrint(("FAILURE! Unable to read Patch Header!\n"));
      return -1;
   }

   return (offset + PATCH_FILE_HEADERINFO_SIZE + pPFInfo->PatchDataLength + gPatchFileCRCSize);
}

/* Returning the filePos at the end of this CPU-patch part.Return  -1 on errors */
static SINT32 PM_ProcessFileInfo(FILE *fptr, SINT32 filePos, tPMPatchFileInfo *pPFInfo)
{
   UINT8  tempbytes[PATCH_FILE_HEADERINFO_SIZE], *p = NULL;
   UNINT  PatchFileCRCOffset = 0;
   UINT32 calcpayloadcrc = 0;

   /* Initialize variables */
   memset(tempbytes,0,sizeof(tempbytes));

   /* Read patch header information */
   if( -1 != ProcessPatchHeader(fptr, filePos, pPFInfo))
   {
      /* Read on patch file CRC. */
      memset(tempbytes, 0, gPatchFileCRCSize);
      PatchFileCRCOffset = PATCH_FILE_CONTENT_START_OFFSET + pPFInfo->PatchDataLength;
      if (0 == Read_PD2(fptr, filePos + PatchFileCRCOffset, tempbytes, gPatchFileCRCSize))
      {
         p = tempbytes;

         if(CRC_32 == gCRCType)
         {
            pPFInfo->PatchDataCRC = p[0]<<24 | p[1]<<16 | p[2]<<8 | p[3];
         }
         else
         {
            pPFInfo->PatchDataCRC = p[0]<<8 | p[1];
         }

         memset(&pPFInfo->PatchData[0],0,sizeof(pPFInfo->PatchData));
         if (0 == Read_PD2(fptr, filePos + PATCH_FILE_CONTENT_START_OFFSET, pPFInfo->PatchData, pPFInfo->PatchDataLength))
         {
             calcpayloadcrc = CRC(pPFInfo->PatchData, pPFInfo->PatchDataLength);
             if (calcpayloadcrc != pPFInfo->PatchDataCRC)
             {
                DebugPrint(("Error %d: PD2 patch payload failed CRC check!\n", ERROR_INVALID_CRC));
                SetErrorCode(ERROR_INVALID_CRC);
                return -1;
             }
             else
             {
                DebugPrint(("CRC passed! Calc CRC = 0x%08x, PD2 CRC = 0x%08x\n",calcpayloadcrc, pPFInfo->PatchDataCRC));
             }
         }
         else
         {
            DebugPrint(("Error %d: PM_ProcessFileInfo failed to read PD2 payload data.\n", ERROR_FILE_READ));
            SetErrorCode(ERROR_FILE_READ);
            return -1;
         }
      }
      else
      {
         /* Need to create zero length patch file with just CRC value of 0? */
         DebugPrint(("Error %d: PM_ProcessFileInfo failed to read CRC value.\n", ERROR_FILE_READ));
         SetErrorCode(ERROR_FILE_READ);
         return -1;
      }
   }

   return filePos + PatchFileCRCOffset + gPatchFileCRCSize;
}

static SINT32 IsK32PatchAvailable(FILE *fptr)
{
   tPMPatchFileInfo PFInfo;
   SINT32 TotalARMPatchSize, TotalFileSize;
   SINT32 filepos = -1;

   TotalARMPatchSize = ProcessPatchHeader(fptr, 0, &PFInfo);

   TotalFileSize = PM_GetFileSize(fptr);

   if(TotalARMPatchSize == TotalFileSize)
   {
      /* K32 patch doesn't exist in the patch file */
      DebugPrint(("Only ARM patches exist in the input patch file\n\n"));
      filepos = -1;
   }
   else if (TotalFileSize > TotalARMPatchSize)
   {
      DebugPrint(("Both ARM and K32 patches exist in the input patch file\n\n"));
      filepos = TotalARMPatchSize; /* Start of K32 patch */
   }
   else
   {
      /* We should never come here */
      filepos = -1;
   }

   return filepos;
}

/* Will return how much of the file used, so the next call can increment that file by that much and see if there is additional patches
   for other CPU's in there. Returns -1 for errors 
*/
static SINT32 LoadAndApplyPatch(FILE *pd2, SINT32 filePos, BOOL sendStartReq, BOOL sendEndReq)
{
   tPMPatchState *ps = &patchState;
   UINT8 *packet;
   LONG  actual_size;
   SINT32 retval =  ERROR_NONE;
 
   /* Initialize global patchState */
   InitializePatchState(!sendStartReq);   /* Note, initialize ValidLoadInProgress to TRUE if we don't need to do the start request */

   /* If file is valid, then process it. */
   filePos = PM_ProcessFileInfo(pd2, filePos, &ps->fileinfo);
   if(filePos == -1)
   {
       DebugPrint(("Error %d: Processing Patch File Failed\n",ERROR_FILE_READ));
       SetErrorCode(ERROR_FILE_READ);
       return -1;
   }

   /* Find start of patch RAM code in Patch Data array and recompute CRC */
   /* for both Overlay and new patch code (Non-Overlay) segments.        */
   if (!FindStartOfPatchRAMCode(&ps->fileinfo, ps->fileinfo.PatchData))
   {
       /* No Non-Overlay Section exists only Overlay Section was found */
       DebugPrint(("Only patch overlay data found in patch data\n"));
   }

   /* Send PM Start message */
   if(sendStartReq)
   {
      retval = GeneratePatchManagerStartRequest(&packet, &actual_size, 2);
      if (FAILURE != retval)
      {
         SSB_Send(packet, actual_size);

         retval = ListenRx(response_PMPrompt);

         if (ERROR_NONE != retval)
         {
            DebugPrint(("Error %d: Did not receive PMPrompt response\n", ERROR_PATCH_PM_PROMPT_MSG));
            SetErrorCode(ERROR_PATCH_PM_PROMPT_MSG);
            PM_MemFree(packet);  /* Free packet before returning */
            return -1;
         }

         PM_MemFree(packet);

         /* Only For 5xp ROM 2.1, set K32 Cache Control regsiter 
            to 0 before downloading patch
         */
         if( (patchState.fileinfo.ChipId == 0x50) &&
             (patchState.fileinfo.SiliconVersion == 0x02) &&
             (patchState.fileinfo.ROMVersionCode == 0x0505)
            )
         {
            tSIRF_MSG_SSB_TRKR_PEEKPOKE_CMD msg;
            msg.type    = 1; /* 4-byte poke in tracker space*/
            msg.access  = 4; /* 32-bit (word) access */
            msg.address = 0xFFFF8200; /* K32 cache control register */
            msg.data    = 0x00000000; /* Set the register to value 0 */

#if 0 
            /* used for debugging */
            /* Peek at that address  */
            msg.type    = 0; /* 4-byte peek in tracker space*/
            SendPeekPokeRequest((void *)&msg);
            msg.type    = 1; /* 4-byte poke in tracker space */
#endif

            SendPeekPokeRequest((void *)&msg);

            retval = ListenRx(response_PeekPoke);

            if (ERROR_NONE == retval)
            {
               /* Peek at that address to make sure it's correctly set to 0 */
               msg.type    = 0; /* 4-byte peek in tracker space*/
               SendPeekPokeRequest((void *)&msg);

               retval = ListenRx(response_PeekPoke);
               if (ERROR_NONE != retval)
               {
                  DebugPrint(("Error %d: Did not receive Peek response\n",ERROR_PATCH_PM_PEEK_MSG));
                  SetErrorCode(ERROR_PATCH_PM_PEEK_MSG);
                  return -1;
               }

               /* Set up Tbus Window3 to access upper 4KB of K32 patch RAM from ARM */
               msg.type    = 1; /* 4-byte poke in tracker space*/
               msg.address = 0xFFFFF664; /* TBUS VMIN3 window base address*/
               msg.data    = 0xc0008000; /* Set the register to value 0x8000 offset in K32 PM region */
               SendPeekPokeRequest((void *)&msg);

               retval = ListenRx(response_PeekPoke);
               if (ERROR_NONE == retval)
               {
                  msg.address = 0xFFFFF66C; /* TBUS VMIN3 window config*/
                  msg.data    = 0x00000063; /* Enable read/write with 4KB window size */
                  SendPeekPokeRequest((void *)&msg);
                  retval = ListenRx(response_PeekPoke);
               }
            }

            if (ERROR_NONE != retval)
            {
               DebugPrint(("Error %d: Did not receive Poke response\n",ERROR_PATCH_PM_POKE_MSG));
               SetErrorCode(ERROR_PATCH_PM_POKE_MSG);
               return -1;
            }
         }
      }
   }

   /* Use global to start PM load process. */
   /* This global flag only gets set if a valid prompt msg was rcvd. */
   if (ps->ValidLoadInProgress)
   {
      BOOL ok;

      ok = PM_FirstSegment(&ps->fileinfo, &ps->filestate);

      /* Change load state */
      ps->loadstate = PM_LOAD_INPROGRESS;

      while(ok)
      {
         /* Re-initialize actual_size variable */
         actual_size = 0;

         if (FAILURE != GeneratePatchManagerLoadRequest(&packet, &actual_size, ps->filestate.seq, ps->filestate.len, ps->fileinfo.PatchData))
         {
            SSB_Send(packet, actual_size);

            ListenRx(response_PMAck);  /* Sets ack_rcvd flag if successful. */

            PM_MemFree(packet);
         }
         else
         {
            /* no data to send or failure to allocate memory */ 
            ok = FALSE;
         }

         if (ok)
         {
             if (ps->filestate.ack_rcvd)  /* Rcvd good Ack */
             {
                ok = PM_NextSegment(&ps->filestate);

                if (!ok)
                   ps->loadstate = PM_LOAD_DONE;
             }
             else /* Rcvd bad Ack */
             {
                /* Failed to get good ack, so break. Packet already freed  */
                /* above, so no need to free packet here. This will allow  */
                /* the App to send Exit cmd to the Rx and terminate load   */
                /* sequence when a failed ack is received.                 */
                break;
             }

         } /* if (ok) */

      } /* while(ok) */
   
   } /* if (ps->ValidLoadInProgress) */
   
   /* If retval flag does not indicate FAILURE, then send PM Exit message. */
   if( (FAILURE != retval) && (sendEndReq))
   { 
      retval = GeneratePatchManagerExitRequest(&packet, &actual_size, 2);
      if (FAILURE != retval)
      {
         ps->filestate.ack_rcvd = FALSE;

         SSB_Send(packet, actual_size);

         retval = ListenRx(response_PMAck);

         if (ERROR_NONE != retval)
         {
            DebugPrint(("ERROR: Did not receive a successful PMExit Ack response\n"));
            PM_MemFree(packet);  /* Free packet before returning */
            return -1;
         }

         PM_MemFree(packet);
      }
   }

   return filePos;
}

static void InitializePatchState(BOOL validLoadInProgress)
{
   memset(&patchState,0,sizeof(tPMPatchState));
   patchState.loadstate = PM_LOAD_IDLE;
   patchState.filestate.seq = 0;
   patchState.ValidLoadInProgress = validLoadInProgress;
}

static BOOL PM_FirstSegment(tPMPatchFileInfo *pi, tPMPatchFileState *pf)
{
   UNINT left;
   BOOL  send_data = FALSE;

   /* This code sends overlay section in the first segment and then sends the */
   /* new patch data segments in subsequent segment(s).                       */
   /* Compute state fields for patch file, which may be zero length */
   if (pi->PatchRAMStartOffset)
   {
      left = pi->PatchRAMStartOffset + gPatchFileCRCSize;
   }
   else
   {
      left = 0;
   }
   /* Initial patch file sequence number is always one */
   pf->seq = 1;
   /* Starting offset into the patch data array */
   pf->offs = 0;
   /* Leave total number of bytes in the patch data plus the patch data crc */
   if (pi->PatchDataLength)
      pf->total = pi->PatchDataLength + gPatchFileCRCSize;
   else
      pf->total = 0;
   /* Initial Patch Load Header Size is 14 bytes instead of 16 bytes. */
   pf->hlen = PATCHLOAD_HDRSZ;
   /* Set 1st load message length based on the minimum between the load */
   /* message size or how much patch data is left.                      */
   if (pi->PatchRAMStartOffset > 0)
   {
      pf->len = MIN(PM_MAX_LOADDATA_SIZE - pf->hlen, left);
   }
   else
   {
      pf->len = MIN(PM_MAX_LOADDATA_SIZE - pf->hlen, pf->total);
   }
   /* Decided not to allow a zero length patch to be sent. Previously for 3tw */
   /* a zero length patch could be sent to delete current patch. */
   if (pf->len > 0)
   {
      send_data = TRUE;
      pf->ack_rcvd = FALSE;
   }

   return send_data;
}

static BOOL PM_NextSegment(tPMPatchFileState *pf)
{
   UNINT left;
   BOOL  send_data = FALSE;
   BOOL  good_flag = FALSE;

   /* Add previous sent patch load message length to running patch data array offset value */
   if ((pf->seq == 1) && (pf->len >= gPatchFileCRCSize))
   {
      if (patchState.fileinfo.PatchRAMStartOffset > 0)
      {
         pf->offs = pf->offs + (pf->len - gPatchFileCRCSize);
         if (pf->len == patchState.fileinfo.PatchRAMStartOffset + gPatchFileCRCSize)
            good_flag = TRUE; 
      }
      else
      {
         pf->offs = pf->offs + pf->len;
         if (pf->total == patchState.fileinfo.PatchDataLength + gPatchFileCRCSize)
            good_flag = TRUE;
      }
   }
   else
      pf->offs += pf->len;
   /* Compute what is left in patch file */
   left = pf->total - pf->offs;
   /* Only first PM load message has additional header info */
   /* PM_MAX_LOADDATA_SIZE removes msg id, sub id and seq no already */
   pf->hlen = 0;
   /* Set minimum length to be sent */
   pf->len = MIN(PM_MAX_LOADDATA_SIZE, left);
   /*  Condition for last load message to be sent
   left = pf->total - (pf->offs + pf->len)
   if (left <= PM_MAX_LOADDATA_SIZE)
   */
   /* If something to send, send it */
   if (pf->len > 0)
   {
      pf->seq = pf->seq + 1;
      send_data = TRUE;
      pf->ack_rcvd = FALSE;
   }

   return send_data;
}

static SINT32 GeneratePatchManagerStartRequest(UINT8 **ppData, LONG *plSize, LONG lNumBytesToWrite)
{
    UINT8 *pData = NULL;
    SINT32 count=0;
    SINT32 i = 0;

    if(lNumBytesToWrite <=0)
    {
        *plSize = 0;
        DebugPrint(("Error %d: Invalid message size (0)\n",ERROR_PATCH_PM_START_MSG));
        SetErrorCode(ERROR_PATCH_PM_START_MSG);
        return FAILURE;
    }

    /* Generate buffer */
    pData = (UINT8*)PM_MemAlloc(sizeof (UINT8) * (lNumBytesToWrite + 100));

    if(NULL == pData)
    {
        DebugPrint(("Error %d: Error in Memory Allocation\n",ERROR_MEMALLOC));
        SetErrorCode(ERROR_MEMALLOC);
        return FAILURE;
    }

    /* Make sure newly create memory is zeroed before using it. */
    PM_MemSet(pData, 0, lNumBytesToWrite + 100);

    /* This message has msg id and sub id only, no other information */
    pData[count++] = MID_SWToolBox;        /* 0xB2 */
    pData[count++] = SID_PM_START;         /* 0x28 */

    *ppData = pData;
    *plSize = count++;   /* Returns correct count because count is incremented after assignment. */

    return ERROR_NONE;
}

static SINT32 GeneratePatchManagerLoadRequest(UINT8 **ppData, LONG *plSize, UINT16 seqNum, LONG lNumBytesToWrite, UINT8 *pDataToWrite)
{
    UINT8 *pData = NULL;
    SINT32 count=0;
    SINT32 i = 0;
    tPMPatchState *ps = &patchState;
    tPMPatchFileInfo *pi = &ps->fileinfo;
    tPMPatchFileState *pf = &ps->filestate;
    UNINT  left;
    UINT32 calcpayloadcrc;

    /* Make sure number of bytes to write is not 0 or negative. */
    if(lNumBytesToWrite <=0)
    {
        *plSize = 0;
        DebugPrint(("Error %d: Invalid message size (0)\n",ERROR_PATCH_PM_LOAD_MSG));
        SetErrorCode(ERROR_PATCH_PM_LOAD_MSG);
        return FAILURE;
    }

    /* Generate buffer */
    pData = (UINT8*)PM_MemAlloc(sizeof (UINT8) * (lNumBytesToWrite + 100));

    if(NULL == pData)
    {
        DebugPrint(("Error %d: Error in Memory Allocation\n",ERROR_MEMALLOC));
        SetErrorCode(ERROR_MEMALLOC);
        return FAILURE;
    }

    /* Make sure newly create memory is zeroed before using it. */
    PM_MemSet(pData, 0, lNumBytesToWrite + 100);

    /* Initial header */
    pData[count++] = MID_SWToolBox;
    pData[count++] = SID_PM_LOAD;
    count = writeSeqNumToStream(pData,count,seqNum);

    if (seqNum == 1 ) 
    {
       UINT16 actual_pdl = pi->PatchDataLength;
       UINT16 actual_prso = pi->PatchRAMStartOffset;
       pData[count++] = 'P';
       pData[count++] = 'M';
       pData[count++] = (UINT8)(pi->ROMVersionCode>>8 & 0xff);
       pData[count++] = (UINT8)(pi->ROMVersionCode & 0xff);
       pData[count++] = (UINT8)(pi->PatchRevisionCode>>8 & 0xff);
       pData[count++] = (UINT8)(pi->PatchRevisionCode & 0xff);
       if(!gK32PatchFlag)
       {
          pData[count++] = (UINT8)((pi->PatchDataBaseAddress&0xFF000000)>>24);
          pData[count++] = (UINT8)((pi->PatchDataBaseAddress&0x00FF0000)>>16);
          pData[count++] = (UINT8)((pi->PatchDataBaseAddress&0x0000FF00)>>8);
          pData[count++] = (UINT8)((pi->PatchDataBaseAddress&0x000000FF));
       }

       /* Add CRC size to patch data length */
       /* Make sure 4e remembers that the patch data length is gPatchFileCRCSize extra bytes, so */
       /* actual patch data is - gPatchFileCRCSize bytes */
       actual_pdl = actual_pdl + gPatchFileCRCSize;
       pData[count++] = (UINT8)(actual_pdl>>8 & 0xff);
       pData[count++] = (UINT8)(actual_pdl & 0xff);

       if(!gK32PatchFlag)
       {
          /* Add CRC size to PatchRAMStartOffset only if PatchRAMStartOffset is > 0 */
          /* Make sure 4e remembers that if the PatchRAMStartOffset is > 0 that it  */
          /* contains gPatchFileCRCSize extra bytes for the overlay CRC.            */
          if (actual_prso > 0)
              actual_prso = actual_prso + gPatchFileCRCSize;
          /* Note PatchRAMStartOffset of 0 means that there is no non-overlay */
          /* section.                                                         */        
          pData[count++] = (UINT8)((actual_prso>>8) & 0xff);
          pData[count++] = (UINT8)(actual_prso & 0xff);
       }
    }
    /* Determine if patch CRC value is included in this load segment */
    /* Compute what is left in patch file */
    left = pf->total - (pf->offs + pf->len); 
    
    /* If final load segment, then don't forget to add patch CRC */
    if (left == 0)
    {
       /* Last load segment */
       /* Copy patch data */
       for(i=0; i<(lNumBytesToWrite-gPatchFileCRCSize); i++)
       {
           pData[count++] = (UINT8) pDataToWrite[pf->offs+i];
       }
       if (pi->PatchRAMStartOffset > 0)
       {
          /* Patch NonOverlay CRC */

          /* Check if PatchRevisionCode is 255 */
          if (pi->PatchRevisionCode == 0xff)
          {
             /* PatchRevisionCode will be used to delete */
             /* existing patch.                          */
             calcpayloadcrc = ~(pi->PatchDataNOVLCRC);
          }
          else
          {
             calcpayloadcrc = pi->PatchDataNOVLCRC;
          }
       }
       else
       {
          /* Patch Overlay CRC */

          /* Check if PatchRevisionCode is 255 */
          if (pi->PatchRevisionCode == 0xff)
          {
             /* PatchRevisionCode will be used to delete */
             /* existing patch.                          */
             calcpayloadcrc = ~(pi->PatchDataOVLCRC);
          }
          else
          {
             calcpayloadcrc = pi->PatchDataOVLCRC;
          }
       }

       if(CRC_32 == gCRCType)
       {
          pData[count++]=(UINT8)((calcpayloadcrc&0xFF000000)>>24);
          pData[count++]=(UINT8)((calcpayloadcrc&0x00FF0000)>>16);
          pData[count++]=(UINT8)((calcpayloadcrc&0x0000FF00)>>8);
          pData[count++]=(UINT8)((calcpayloadcrc&0x000000FF));
       }
       else
       {
          pData[count++]=(UINT8)((calcpayloadcrc&0x0000FF00)>>8);
          pData[count++]=(UINT8)((calcpayloadcrc&0x000000FF));
       }
    }
    else
    {
       LONG lDataBytesToWrite = 0;

       /* Not last load segment */
       if (seqNum == 1)
       {
          if (lNumBytesToWrite >= gPatchFileCRCSize)
             lDataBytesToWrite = lNumBytesToWrite - gPatchFileCRCSize;
       }
       else
       {
          lDataBytesToWrite = lNumBytesToWrite;
       }

       for(i=0; i<lDataBytesToWrite; i++)
       {
           pData[count++] = (UINT8) pDataToWrite[pf->offs+i];
       }

       /* Calculate overlay CRC for 1st load message only */
       if (seqNum == 1)
       {
          /* Patch data CRC for overlay section only */

          /* Check if PatchRevisionCode is 255 */
          if (pi->PatchRevisionCode == 0xff)
          {
             /* PatchRevisionCode will be used to delete */
             /* existing patch.                          */
             calcpayloadcrc = ~(pi->PatchDataOVLCRC);
          }
          else
          {
             calcpayloadcrc = pi->PatchDataOVLCRC;
          }
          if(CRC_32 == gCRCType)
          {
             pData[count++]=(UINT8)((calcpayloadcrc&0xFF000000)>>24);
             pData[count++]=(UINT8)((calcpayloadcrc&0x00FF0000)>>16);
             pData[count++]=(UINT8)((calcpayloadcrc&0x0000FF00)>>8);
             pData[count++]=(UINT8)((calcpayloadcrc&0x000000FF));
          }
          else
          {
             pData[count++]=(UINT8)((calcpayloadcrc&0x0000FF00)>>8);
             pData[count++]=(UINT8)((calcpayloadcrc&0x000000FF));
          }
       }
    }

    *ppData = pData;
    *plSize = count++;

    return ERROR_NONE;
}

static SINT32 GeneratePatchManagerExitRequest(UINT8 **ppData, LONG *plSize, LONG lNumBytesToWrite)
{
    UINT8 *pData = NULL;
    SINT32 count=0;
    SINT32 i = 0;

    if(lNumBytesToWrite <=0)
    {
        *plSize = 0;
        DebugPrint(("Error %d: Invalid message size (0)\n",ERROR_PATCH_PM_EXIT_MSG));
        SetErrorCode(ERROR_PATCH_PM_EXIT_MSG);
        return FAILURE;
    }

    /* Generate buffer */
    pData = (UINT8*)PM_MemAlloc(sizeof (UINT8) * (lNumBytesToWrite + 100));

    if(NULL == pData)
    {
        DebugPrint(("Error %d: Error in Memory Allocation\n",ERROR_MEMALLOC));
        SetErrorCode(ERROR_MEMALLOC);
        return FAILURE;
    }

    /* Make sure newly create memory is zeroed before using it. */
    PM_MemSet(pData, 0, lNumBytesToWrite + 100);

    /* This message has msg id and sub id only, no other information */
    pData[count++] = MID_SWToolBox;        /* 0xB2 */
    pData[count++] = SID_PM_EXIT;          /* 0x26 */

    *ppData = pData;
    *plSize = count++;   /* Returns correct count because count is incremented after assignment. */

    return ERROR_NONE;
}

static SINT32 writeSeqNumToStream(UINT8 *pData, SINT32 count, UINT16 seqNum)
{
   if(gK32PatchFlag)
   {
      seqNum |= (1<<15);   /* set MSB if it's a K32 patch */
   }
   pData[count++] = (UINT8)((seqNum&0xFF00)>>8);
   pData[count++] = (UINT8)((seqNum&0x00FF));
   return count;   
}

/* Processes the patch payload data from the pd2 file to find the start of */
/* Patch RAM code (non-overlay section) if it exist. Computes CRC values */
/* for both overlay and non-overlay sections. Returns TRUE if Patch RAM    */
/* code is found, else returns FALSE.                                      */
static BOOL FindStartOfPatchRAMCode(tPMPatchFileInfo *pPFInfo, UINT8 *patchdata)
{
   UINT8 *base,*p;
   UNINT len = pPFInfo->PatchDataLength;
   UNINT cur = 0;
   UNINT overlaysect = 0;
   UNINT segoff, seglen;
   BOOL  return_flag = FALSE;
   BOOL  nonoverlay_flag = FALSE;
  
   /* Process patch payload data until cur equal len */
   while (cur < len)
   {
      base = p = patchdata + cur;
      /* Check if segment offset (addr) is an overlay. */
      if(!gK32PatchFlag)
      {
         segoff = PATCH_READRAW(p, UINT32);
         if (segoff < 0x200000)
         {
            /* Overlay section, so 1st byte is overlay type and 2nd byte is segment length. */
            /*overtype = */PATCH_READRAW(p, UINT8);
            seglen = PATCH_READRAW(p, UINT8);
            overlaysect++;
         }
         else
         {
            /* For new patch function code, segment length is 2 bytes like before */
            seglen = PATCH_READRAW(p, UINT16);
            pPFInfo->PatchRAMStartOffset = cur;
            nonoverlay_flag = TRUE;
            /* If pPFInfo->PatchRAMStartOffet equals pPFInfo->PatchDataLength   */
            /* implies patch only contains overlays and no new patch functions. */
         }
      }
      else
      {  /* For K32 patches we always have 1 overlay section and 1 patch section, both indicated by a 16bit length */
         seglen = PATCH_READRAW(p, UINT16);
         if(cur == 0)   /* First section always overlay */
         {
            overlaysect++;
         }
         else  /* Second section is always the patch data */
         {
            pPFInfo->PatchRAMStartOffset = cur;
            nonoverlay_flag = TRUE;
         }
      }
      cur += p - base + seglen;
   }
   /* If overlay sections found, then process them. */
   if (overlaysect > 0)
   {
      /* Calculate CRC for patch overlay section only */
      if (pPFInfo->PatchRAMStartOffset > 0)
      {
         if(!gK32PatchFlag)
            pPFInfo->PatchDataOVLCRC = CRC(patchdata, pPFInfo->PatchRAMStartOffset);
         else
            pPFInfo->PatchDataOVLCRC = CRC(patchdata+2, pPFInfo->PatchRAMStartOffset-2);

      }
      else
      {
         /* Case where pPFInfo->PatchRAMStartOffet equals pPFInfo->PatchDataLength */
         pPFInfo->PatchDataOVLCRC = CRC(patchdata, pPFInfo->PatchDataLength);
      }
   }
   /* Print out Overlay Section CRC */
   DebugPrint(("Overlay Section CRC = 0x%08x\n",pPFInfo->PatchDataOVLCRC));
   
   /* If Non-Overlay section found, then compute its CRC without its segment address and segment length. */
   if (nonoverlay_flag)
   {
      /* Calculate CRC for patch nonoverlay section */
      /* Needed to add 6 to not perform CRC on offset and byte size for nonoverlay section */
      if(gK32PatchFlag)
      {
         pPFInfo->PatchDataNOVLSize = pPFInfo->PatchDataLength - pPFInfo->PatchRAMStartOffset;
         pPFInfo->PatchDataNOVLCRC = CRC(patchdata + pPFInfo->PatchRAMStartOffset + 2, 
                                        pPFInfo->PatchDataNOVLSize - 2);
      }
      else
      {
         pPFInfo->PatchDataNOVLSize = pPFInfo->PatchDataLength - pPFInfo->PatchRAMStartOffset;
         pPFInfo->PatchDataNOVLCRC = CRC(patchdata + pPFInfo->PatchRAMStartOffset + 6, 
                                        pPFInfo->PatchDataNOVLSize - 6);
      }
      /* Set return flag to TRUE to indicate PatchRAMStartOffset was found */
      return_flag = TRUE;
  
   }
   else
   {
      /* Case where NO NonOverlay Section was found in patch. */
      pPFInfo->PatchDataNOVLCRC = 0;
      pPFInfo->PatchDataNOVLSize = 0;
   }
   /* Print out Non-Overlay Section CRC */
   DebugPrint(("Non-Overlay Section CRC = 0x%08x\n",pPFInfo->PatchDataNOVLCRC));

   /* If TRUE is returned, then the start of Patch RAM code was found meaning */
   /* Non-Overlay Section was found.                                          */
   return return_flag;
   
}

/**
 * @brief      Patch_readRaw reads a specific number of bytes from the 
 *             buffer provided
 * @param[in]  Address of the buffer, the bytes to read
 * @return     Value read
 */
static UINT32 Patch_readRaw (UINT8 *pRaw, UINT32 size)
{
   UINT32 v = 0;
   while (size-- > 0)
   {
      v <<= 8;
      v += *pRaw++;
   }
   return v;
}

static UINT8 GetPowerMode(RcvrParamSS_2 *localParam)
{
    UINT8 flag = 0; /* 0 = full power , 1 = trickle power,2 = PTF */

    if( localParam->LP_PushToFix )
    {
        flag = 2;
    }
    else if( !localParam->LP_PwrCyclingEnabled )
    {
        flag = 0;
    }
    else
    {
        flag = 1;
    }
    return flag;
}
