/*                                                                        *
*    File is Copyright  2013 CSR Technology Inc.  All rights reserved.   *
*    No implied or expressed warranties are granted.                      *
*                                                                         *
*    This Software is protected by United States copyright laws and       *
*    international treaties.  You may not reverse engineer, decompile     *
*    or disassemble this Software.                                        *
*                                                                         *
*    WARNING:                                                             *
*    This Software contains CSR Technology Inc.s confidential and        *
*    proprietary information. UNAUTHORIZED COPYING, USE, DISTRIBUTION,    *
*    PUBLICATION, TRANSFER, SALE, RENTAL OR DISCLOSURE IS PROHIBITED      *
*    AND MAY RESULT IN SERIOUS LEGAL CONSEQUENCES.  Do not copy this      *
*    Software without CSR Technology, Inc.s  express written             *
*    permission. Use of any portion of the contents of this Software      *
*    is subject to and restricted by your signed written agreement with   *
*    CSR Technology, Inc.                                                 *
*                                                                         *
***************************************************************************
*
* MODULE:  ROM Patch Application
*
* FILENAME:  main.c
*
* DESCRIPTION:This file contains the source code for main application and reading the 
*                       input configuration
*
***************************************************************************
*
*  Keywords for Perforce.  Do not modify.
*
*  $File: //core/tools/ROMPatchApplication/src/main.c $
*
*  $DateTime: 2015/11/12 20:51:56 $
*
*  $Revision: #18 $
*
***************************************************************************/

/**************************************************************************/
/*                              INCLUDES                                  */
/**************************************************************************/
#include "Abstraction/basicTypes.h"
#include "ROMPatchSettings.h"
#include "Abstraction/osAbstraction.h"
#include "Abstraction/logs.h"
#include "protocol.h"
#include "errorCodes.h"
#include "Abstraction/CommInterface.h"
#include "ssbInterface.h"
#include "Abstraction/string_abstraction.h"

/**************************************************************************/
/*                          GLOBALS                                       */
/**************************************************************************/
/*************************************/
/* Patch Manager Version Information */
/*************************************/
#define MAJOR_VER          (3)
#define MINOR_VER          (2)
#define MONTH              (11)
#define DAY                (11)
#define YEAR               (2015)
/*************************************/

UINT8  CustomerConfig[NUM_CONFIG][MAX_CONFIG_SIZE];
tInputConfig inputConfiguration;

/**************************************************************************/
/*              STATIC FUNCTION PROTOTYPES                                */
/**************************************************************************/
static SINT32 GetConfig(UINT8 *ptr,LONG size);
static SINT32 SetupSSBCommWithRx(UINT32 protocol, UINT32 baud);
static SINT32 CommErrorRecovery(SINT32 errorCode);
static SINT32 AppInit(SINT32 argc, SINT8 *argv[]);
static void   PrintSummary(void);

/**************************************************************************/
/*                     GLOBAL FUNCTIONS                                   */
/**************************************************************************/
SINT32 main(SINT32 argc, SINT8 *argv[])
{
    SINT32 returnStatus;

    SetErrorCode(ERROR_NONE);

    /* Parse input configuration file and open COM port*/
    returnStatus = AppInit(argc,argv);
    if(ERROR_NONE == returnStatus )
    {
        DebugPrint(("\n------Setup And Verify Communication with Rx-----\n"));
        returnStatus = SetupSSBCommWithRx(inputConfiguration.initialRxProtocol,inputConfiguration.initialRxBaudRate);
    }
    else
    {
       /* Results summary */
       PrintSummary();

       if(ERROR_NONE == CloseLog())
       {
          DebugPrint(("Log File Closed \n"));
       }
       return FAILURE;
    }

    /* Handle if the receiver is sending the data but baud or protocol is mismatched */
    if(ERROR_COM_BYTES_UNRECOGNIZED == returnStatus)
    {
        DebugPrint(("\n------Trying to Recover from Comm Error-----\n"));
        returnStatus = CommErrorRecovery(returnStatus);
    }

    if(ERROR_NONE == returnStatus)
    {
        DebugPrint(("\n------Handling Power Mode -----\n"));
        returnStatus = HandleLPModeScenario();
    }

    /* If receiver is not responding at all*/
    if(ERROR_RCVR_NOT_RESPONDING == returnStatus)
    {
        DebugPrint(("Error %d: FAILURE! Check Receiver is either Switched Off or In Low Power Mode\n",ERROR_RCVR_NOT_RESPONDING));
        SetErrorCode(ERROR_RCVR_NOT_RESPONDING);
    }

    if( (ERROR_NONE == returnStatus) || (ERROR_NONE_SSB_MSG_RCVD == returnStatus))
    {
        DebugPrint(("\n------Process, Load and Apply Patch-----\n"));
        (void) ProcessPatchFile(inputConfiguration.patchFilePath);
    }

    /* If receiver is sending the response which is recognized by the Patch Manager */
    if((ERROR_RCVR_NOT_RESPONDING != returnStatus) && 
       (ERROR_COM_BYTES_UNRECOGNIZED != returnStatus)
      )
    {
        /* Check CCK IO Protocol Flag before restoring Rx's Initial Power Mode */
        /* and Initial Comm Settings.                                             */
        if (!inputConfiguration.CCKIOProtoFlag)
        {
            DebugPrint(("\n------Restoring Initial Power Mode Settings of Rx-----\n"));

            returnStatus = RestoreLPModeSettings();

            DebugPrint(("\n------Restoring Initial Comm Settings of Rx-----\n"));
            returnStatus =  RestoreRxSettings();

            if(ERROR_NONE == returnStatus)
            {
                DebugPrint(("Initial Settings of Receiver Restored\n"));
            }
            else
            {
                DebugPrint(("Failed to Restore Initial Settings \n"));
            }
        }
        else
        {
            DebugPrint(("\n------CCK patch applied, Rx uart settings should match CCK default-----\n"));
        }
    }
    else
    {
        DebugPrint(("Error %d: Failed to communicate with receiver. Patch Will not be applied.\n",ERROR_RCVR_NOT_RESPONDING));
        SetErrorCode(ERROR_RCVR_NOT_RESPONDING);
    }

    /* Close Communication */
    if(ERROR_NONE == Comm_DevClose(NULL))
    {
        DebugPrint(("COM Port Closed\n"));
    }

    /* Results summary */
    PrintSummary();

    PrintLog("\n*********END OF LOG SESSION **********\n\n");
    if(ERROR_NONE == CloseLog())
    {
        DebugPrint(("Log File Closed \n"));
    }

} /* main */

/**************************************************************************/
/*                     STATIC FUNCTIONS                                   */
/**************************************************************************/
static SINT32 AppInit(SINT32 argc, SINT8 *argv[])
{
    LONG size=0;
    UINT8 *ptr;
    UINT8 i=0;
    SINT32 returnStatus = ERROR_NONE;

    if (argc == 2)
    {
        DebugPrint(("%s\n",argv[1]));
        DebugPrint(("------Reading Input Configuration------\n"));
        returnStatus = ReadBinaryFile(argv[1],&ptr,&size);
    }
    else
    {
        DebugPrint(("Error %d: Provide Config File Name as input Argument \n",ERROR_INVALID_INPUT));
        DebugPrint(("Exiting the Application\n"));
        SetErrorCode(ERROR_INVALID_INPUT);
        return ERROR_INVALID_INPUT;
    }

    if(returnStatus != ERROR_NONE)
    {
        DebugPrint(("Error %d: Reading Input Config File\n   Check input arguments \n",ERROR_FILE_OPEN));
        DebugPrint(("Exiting the application...\n"));
        SetErrorCode(ERROR_FILE_OPEN);
        return ERROR_FILE_OPEN;
    }

    DebugPrint(("\n------Verifying Input configuration from User-----\n"));

    if (FAILURE == GetConfig(ptr, size))
    {
        /* Exiting the application due to invalid input */
        PM_MemFree(ptr);
        DebugPrint(("Error %d: Invalid configuration input\n",ERROR_INVALID_INPUT));
        DebugPrint(("Exiting the Application\n"));
        SetErrorCode(ERROR_INVALID_INPUT);
        return ERROR_INVALID_INPUT;
    }
    else
    {
        DebugPrint(("\nInput configuration successfully verified.\n"));
    }

    /* Free the pointer allocated in ReadBinary File */
    PM_MemFree(ptr);

    /* Initialize the logging session */
    if(ERROR_NONE == LogInit(inputConfiguration.logFilePath))
    {
        PrintLog("\n********START OF LOG SESSION *********\n");
        DebugPrint(("Log File %s Opened\n" , inputConfiguration.logFilePath));
    }
    else
    {
        DebugPrint(("Error %d: Failed to Open Log File %s \n" , ERROR_FILE_OPEN, inputConfiguration.logFilePath));
        DebugPrint(("Logs will not be generated\n "));
        //SetErrorCode(ERROR_FILE_OPEN); - Not a catastrophic failure
    }

    DebugPrint(("\n"));
    DebugPrint(("-----------------------------------------------\n"));
    DebugPrint(("------CSR GSD ROM Patch Application------------\n"));
    DebugPrint(("------Version: %d.%d, Date: %d/%d/%d------------\n",MAJOR_VER, MINOR_VER, MONTH, DAY, YEAR));
    DebugPrint(("-----------------------------------------------\n\n"));

    PrintLog("\nPatchMgr Input Configuration :\n ");
    for(i=0;i<NUM_CONFIG;i++)
    {
        PrintLog("%s ",CustomerConfig[i]);
    }

    /************** Opening the COM Port ******************/
    ptr = (UINT8*)PM_MemAlloc(sizeof(UINT8) * 13);
    PM_MemCpy((void*)ptr,(void*)inputConfiguration.portName,10);

    ptr[10] = (inputConfiguration.initialRxBaudRate >> 16) & 0xFF;
    ptr[11] = (inputConfiguration.initialRxBaudRate >> 8) & 0xFF;
    ptr[12] = inputConfiguration.initialRxBaudRate & 0xFF;

    DebugPrint(("\n------Opening COM Port-----\n"));
    if(Comm_DevInit((void*)ptr) == FAILURE)
    {
        DebugPrint(("Error %d: Unable To Open COM Port\n",ERROR_COMPORT_OPEN));
        SetErrorCode(ERROR_COMPORT_OPEN);
        PM_MemFree(ptr);
        return ERROR_COMPORT_OPEN;
    }
    else
    {
        PM_MemFree(ptr);
        DebugPrint(("\nCOM Port Open at %d baudrate(SUCCESS)\n",inputConfiguration.initialRxBaudRate));

        if(NMEA_PROTOCOL == inputConfiguration.initialRxProtocol)
        {
           DebugPrint(("Protocol: NMEA\n"));
        }
        else
        {
           DebugPrint(("Protocol: OSP\n"));
        }
    }

    return returnStatus;
}

static SINT32 GetConfig(UINT8 *ptr,LONG size)
{
    LONG lengthTraversed =0;
    SINT32 configNum=0,i=0;

    PM_MemSet(CustomerConfig,0,MAX_CONFIG_SIZE*NUM_CONFIG);
    while( (lengthTraversed < size) && (configNum < NUM_CONFIG) )
    {
        i=0;
        while((*ptr != ' ') && (lengthTraversed < size) && (*ptr != '\r')&& (*ptr != '\n'))
        {
           if(*ptr == '\"')
            {
               ptr+=1;
               lengthTraversed+=1;
               while((*ptr != '\"') && (lengthTraversed < size) && (*ptr != '\r')&& (*ptr != '\n'))
               {
                  CustomerConfig[configNum][i++] = *ptr++;
                  lengthTraversed+=1;
               }
               ptr+=1;
               lengthTraversed+=1;
            }
            else
            {
               CustomerConfig[configNum][i++] = *ptr++;
               lengthTraversed++;
            }
        }

        ptr+=1;
        lengthTraversed+=1;
        configNum++;
    }

    DebugPrint(("\nPatchMgr Input Configuration :\n "));
    for(i=0;i<NUM_CONFIG;i++)
    {
        DebugPrint(("%s ",CustomerConfig[i]));
    }
    DebugPrint(("\n"));


    for(configNum =0; configNum < NUM_CONFIG; configNum++)
    {
        switch(configNum)
        {
        case 0:
            if(strcmp(CustomerConfig[configNum],"UART") == 0)
                inputConfiguration.portType = UART;
            else if(strcmp(CustomerConfig[configNum],"SPI") == 0)
                inputConfiguration.portType = SPI;
            else
            {
                DebugPrint(("\nError %d: Invalid Port. Enter Valid Value UART or SPI \n",ERROR_INVALID_PORT));
                SetErrorCode(ERROR_INVALID_PORT);
                return FAILURE;
            }
            break;
        case 1:
            PM_strcpy(inputConfiguration.portName, CustomerConfig[configNum]);
            break;
        case 2:
            if(strcmp(CustomerConfig[configNum],"4800") == 0)
                inputConfiguration.initialRxBaudRate = 4800;
            else if(strcmp(CustomerConfig[configNum],"9600") == 0)
                inputConfiguration.initialRxBaudRate = 9600;
            else if(strcmp(CustomerConfig[configNum],"19200") == 0)
                inputConfiguration.initialRxBaudRate = 19200;
            else if(strcmp(CustomerConfig[configNum],"38400") == 0)
                inputConfiguration.initialRxBaudRate = 38400;
            else if(strcmp(CustomerConfig[configNum],"57600") == 0)
                inputConfiguration.initialRxBaudRate = 57600;
            else if(strcmp(CustomerConfig[configNum],"115200") == 0)
                inputConfiguration.initialRxBaudRate = 115200;
            else if(strcmp(CustomerConfig[configNum],"230400") == 0)
                inputConfiguration.initialRxBaudRate = 230400;
            else if(strcmp(CustomerConfig[configNum],"460800") == 0)
                inputConfiguration.initialRxBaudRate = 460800;
            else
            {
                DebugPrint(("\nError %d: Invalid Initial Rcvr Baud Rate. Enter Valid Value 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800\n",ERROR_INVALID_BAUD));
                SetErrorCode(ERROR_INVALID_BAUD);
                return FAILURE;
            }
            break;
        case 3:
            if(strcmp(CustomerConfig[configNum],"OSP") == 0)
                inputConfiguration.initialRxProtocol = OSP_PROTOCOL;
            else if(strcmp(CustomerConfig[configNum],"NMEA") == 0)
                inputConfiguration.initialRxProtocol = NMEA_PROTOCOL;
            else
            {
                DebugPrint(("\nError %d: Invalid Initial Protocol. Enter Valid Value OSP or NMEA\n",ERROR_INVALID_PROTOCOL));
                SetErrorCode(ERROR_INVALID_PROTOCOL);
                return FAILURE;
            }
            break;
        case 4:
            if(strcmp(CustomerConfig[configNum],"4800") == 0)
                inputConfiguration.patchBaudRate = 4800;
            else if(strcmp(CustomerConfig[configNum],"9600") == 0)
                inputConfiguration.patchBaudRate = 9600;
            else if(strcmp(CustomerConfig[configNum],"19200") == 0)
                inputConfiguration.patchBaudRate = 19200;
            else if(strcmp(CustomerConfig[configNum],"38400") == 0)
                inputConfiguration.patchBaudRate = 38400;
            else if(strcmp(CustomerConfig[configNum],"57600") == 0)
                inputConfiguration.patchBaudRate = 57600;
            else if(strcmp(CustomerConfig[configNum],"115200") == 0)
                inputConfiguration.patchBaudRate = 115200;
            else if(strcmp(CustomerConfig[configNum],"230400") == 0)
                inputConfiguration.patchBaudRate = 230400;
            else if(strcmp(CustomerConfig[configNum],"460800") == 0)
                inputConfiguration.patchBaudRate = 460800;
            else
            {
                DebugPrint(("\nError %d: Invalid Patch Baud Rate. Enter Valid Value 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800\n",ERROR_INVALID_BAUD));
                SetErrorCode(ERROR_INVALID_BAUD);
                return FAILURE;
            }
            break;
        case 5:
            /* Replaced this field for 4e ROM Patch Application, since 4e Patch Manage   */
            /* code does an internal cold start after a new patch is loaded.  The 4e ROM */
            /* Patch Application does not issue a software or hardware reset like LT.    */
            if (strcmp(CustomerConfig[configNum],"TRUE") == 0)
               inputConfiguration.CCKIOProtoFlag = TRUE;
            else if (strcmp(CustomerConfig[configNum],"FALSE") == 0) 
               inputConfiguration.CCKIOProtoFlag = FALSE;
            else
            {
                DebugPrint(("\nError %d:Invalid CCK IO Protocol Flag Entered. Enter Valid Flag must be TRUE or FALSE\n",ERROR_INVALID_CCK_FLAG));
                SetErrorCode(ERROR_INVALID_CCK_FLAG);
                return FAILURE;
            }
            break;
        case 6:
            PM_strcpy(inputConfiguration.patchFilePath, CustomerConfig[configNum]);
            break;
        case 7:
            PM_strcpy(inputConfiguration.logFilePath, CustomerConfig[configNum]);
            break;
        default:
            {
                DebugPrint(("\nError %d: Invalid  Number of Arguments\n",ERROR_INVALID_NUM_PARAMS));
                SetErrorCode(ERROR_INVALID_NUM_PARAMS);
            }
            return FAILURE;
        }
    }     
    return ERROR_NONE;
}

static SINT32 SetupSSBCommWithRx(UINT32 protocol, UINT32 baud)
{
    UINT8 count= 0;
    SINT32 returnStatus = ERROR_NONE;
    UINT32 initBaud = baud;
    UINT32 baudRate = inputConfiguration.patchBaudRate;

    /* First just modify baud rate keeping initial protocol same */
    if((NMEA_PROTOCOL== inputConfiguration.initialRxProtocol) && (initBaud != baudRate))
    {
       DebugPrint(("Switch to NMEA baud %d\n",baudRate));
       count++;
       do
       {
          DebugPrint(("Trial %d\n",count));
          returnStatus = SwitchNmeaBaud(baudRate);
          DelayMs(DELAY_AFTER_BAUDRATECHANGE * 20);
       }while((count++ < MAX_NUM_TRIALS) && (returnStatus != ERROR_NONE));

       if(returnStatus == ERROR_NONE)
       {
          DebugPrint(("Successfully switched to NMEA baud %d\n",baudRate));
          initBaud = baudRate;
       }
       else
       {
          DebugPrint(("Switch to NMEA baud %d NOT successful\n",baudRate));
       }
       count = 0;
    }

    DebugPrint(("Trying At BaudRate = %d \n",initBaud));

    //DebugPrint(("Attempt to Receive a Valid OSP Message at %d\n",baudRate));
    while((count++ < MAX_NUM_TRIALS) && (returnStatus != ERROR_NONE_SSB_MSG_RCVD))
    {
        /* Switch to Initial Baud Rate */
        DebugPrint(("Trial %d\n",count));
        Comm_DevIOCtrl(UARTSetBaudrate,(void*)&initBaud);
        DelayMs(DELAY_AFTER_BAUDRATECHANGE);
        if(NMEA_PROTOCOL== protocol)
        {
            returnStatus = SwitchNmeaToSsb(baudRate);
        }
        else
        {
            returnStatus = SetSerialBitRate(baudRate);
        }
        returnStatus = ListenRx(response_recvSSB);

        /* If receiver is not responding for more than 2 wait loops*/
        if((ERROR_RCVR_NOT_RESPONDING == returnStatus) && (count >= 2))
            break;
    }

    if(ERROR_NONE_SSB_MSG_RCVD == returnStatus)
    {
        DebugPrint(("Initial Commn Established..(SUCCESS)\n"));
        returnStatus = ERROR_NONE;
    }
    else if(ERROR_RCVR_NOT_RESPONDING == returnStatus)
    {
        DebugPrint(("Error %d: FAILURE in setting up initial Communication with the receiver \n",ERROR_RCVR_NOT_RESPONDING));
        SetErrorCode(ERROR_RCVR_NOT_RESPONDING);
        returnStatus = ERROR_RCVR_NOT_RESPONDING;
    }
    else
    {
       DebugPrint(("Error %d: FAILURE in setting up initial Communication with the receiver \n",ERROR_COM_BYTES_UNRECOGNIZED));
       SetErrorCode(ERROR_COM_BYTES_UNRECOGNIZED);
       returnStatus = ERROR_COM_BYTES_UNRECOGNIZED;
    }

    return returnStatus;
}

static SINT32 CommErrorRecovery(SINT32 errorCode)
{
    SINT32 returnStatus=ERROR_NONE;
#define NUM_BAUD 8
    UINT32 baudRateArray[NUM_BAUD] = {4800,9600,19200,38400,57600,115200, 230400, 460800};
    UINT8 baudTrial=0;
    UINT32 protocol= OSP_PROTOCOL;

    /* If bytes are received over serial link but received bytes are not recognized  Then 
    First try with different protocol 
    Second try all baud rates and protocol (OSP/NMEA) combination one by one */

    if(errorCode == ERROR_COM_BYTES_UNRECOGNIZED)
    {
        /* Reset error code from previous trial */
        SetErrorCode(ERROR_NONE);

        /* Revert Application back to initial baud rate and try with different protocol
        than what is mentioned in config file */
        DebugPrint(("Trying Other Protocol \n"));
        returnStatus = SetupSSBCommWithRx((inputConfiguration.initialRxProtocol == OSP_PROTOCOL)? NMEA_PROTOCOL:OSP_PROTOCOL, 
            inputConfiguration.initialRxBaudRate);
        if(returnStatus == ERROR_NONE)
        {
            DebugPrint(("Initial Protocol Mentioned is incorrect\n"));
        }
        else if(returnStatus == ERROR_RCVR_NOT_RESPONDING)
        {
            DebugPrint(("FAILURE! Receiver not responding\n"));
        }
        else /*if(returnStatus == ERROR_COM_BYTES_UNRECOGNIZED)*/
        {
            DebugPrint(("Trying different protocol and baud rate combination, will take some time\n"));
            /* Now try with other initial baud rates assuming protocol mentioned is correct*/
            while( (returnStatus == ERROR_COM_BYTES_UNRECOGNIZED) && (protocol <  NMEA_PROTOCOL+1))
            {
                if(protocol == OSP_PROTOCOL)
                {
                    DebugPrint(("\n----Protocol OSP----\n "));
                }
                else
                {
                    DebugPrint(("\n----Protocol NMEA----\n"));
                }
                baudTrial = 0;  
                while( (returnStatus == ERROR_COM_BYTES_UNRECOGNIZED) && (baudTrial < NUM_BAUD))
                {
                   /* Reset error code from previous trial */
                   SetErrorCode(ERROR_NONE);

                   /* Trying At different Baud Rates */
                    returnStatus = SetupSSBCommWithRx(protocol, baudRateArray[baudTrial]);
                    /* returnStatus = VerifyCommunication(); */
                    if(returnStatus == ERROR_NONE)
                    {
                        DebugPrint(("BaudRate %d Pass \n\n",baudRateArray[baudTrial]));
                    }
                    else if(returnStatus == ERROR_RCVR_NOT_RESPONDING)
                    {
                        DebugPrint(("FAILURE! Receiver not responding\n"));
                    }
                    else /*if(returnStatus == ERROR_COM_BYTES_UNRECOGNIZED)*/
                    {
                        DebugPrint(("BaudRate %d did not work \n\n",baudRateArray[baudTrial]));
                    }

                    baudTrial++;
                }
                protocol++;
            }
        }
    }
    else
        returnStatus = errorCode;

    if(ERROR_NONE == returnStatus )
    {
        DebugPrint(("Communication Error Recovered.(SUCCESS)\n"));
    }
    else
    {
        DebugPrint(("Error %d: FAILURE ! Communication Error Recovery Failed\n", returnStatus));
        SetErrorCode(returnStatus);
    }

    return returnStatus;
}

void PrintSummary(void)
{
   DebugPrint(("\n*********RESULT*********\n"));
   if(ERROR_NONE == GetErrorCode())
   {
      DebugPrint(("Summary: SUCCESS - Patch load was successful.\n"));
   }
   else
   {
      DebugPrint(("Summary: FAILURE - Patch load was not successful. Error Code %d. \n", GetErrorCode()));
      printf("Check Error Log for more details.\n\n");
   }
}
